"
 COPYRIGHT (c) 1992 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#ObjectMemory
	instanceVariableNames:''
	classVariableNames:'InternalErrorHandler UserInterruptHandler TimerInterruptHandler
		SpyInterruptHandler StepInterruptHandler
		ExceptionInterruptHandler ErrorInterruptHandler
		MemoryInterruptHandler SignalInterruptHandler
		ChildSignalInterruptHandler DisposeInterruptHandler
		RecursionInterruptHandler IOInterruptHandler
		CustomInterruptHandler RegisteredErrorInterruptHandlers
		InterruptLatencyMonitor AllocationFailureSignal
		MallocFailureSignal LowSpaceSemaphore IncrementalGCLimit
		FreeSpaceGCLimit FreeSpaceGCAmount BackgroundCollectProcess
		BackgroundFinalizationProcess FinalizationSemaphore Dependents
		ImageName ImageSaveTime ChangeFileName MaxInterruptLatency
		InterruptLatencyGoal'
	poolDictionaries:''
	category:'System-Support'
!

!ObjectMemory class methodsFor:'documentation'!

caching
"
    The system uses various caches to speed up method-lookup.
    Currently, there is a three-level cache hierarchy:

        inline-cache            keeps the target of the last send at the caller-
                                side (i.e. every send goes through its private 
                                1-slot inline-cache, where the address of the last
                                called function at this call location is kept.)

        polymorph-inline-cache  keeps a limited list of all targets ever reached 
                                at this call location. The list is automatically 
                                flushed if it grows too large, or the overall number
                                of poly-chache entries exceeds a limit.

        method-lookup-cache     a global cache. Hashes on class-selector pairs,
                                returning the target method.

    Whenever methods are added or removed from the system, or the inheritance 
    hierarchy changes, some or all caches have to be flushed.
    The flushXXX methods perform the task of flushing various caches.
    All standard methods in Behavior call for cache flushing, when things change;
    however, if you use the low level access methods in Behavior
    (for example: #setSuperclass:) special care has to be taken.

    In some situations, not all caches need flushing, for example a change
    in an interpreted method (currently) needs no flushing of the inline caches.
    Also, flushing can be limited to entries for a specific class for most changes.

    To be 'on the brigth side of live', use ObjectMemory>>flushCaches (which
    flushes all of them), when in doubt of which caches should be flushed. 
    It is better flush too much - otherwise you may end up in a wrong method after 
    a send (you wont really see a difference in speed after the flush, anyway).
"
!

copyright
"
 COPYRIGHT (c) 1992 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    This class contains access methods to the system memory and the VM.

    In previous ST/X versions, this stuff used to be in the Smalltalk class.
    It has been separated for better overall class structure and modularisation.
    There are no instances of ObjectMemory - all is done in class methods.
    (this is a functional interface).

    Many methods here are for debuging purposes, for developers
    or experimental, and therefore not standard.
    Do not depend on them being there - some may vanish ...
    (especially those, that depend on a specific GC implementation)
    Most of the stuff found here is not available, or different or called
    different in other smalltalk implementations. Be aware, that using these
    interfaces (especially: depending on them) may make your application
    non portable.

    See more documentation in -> caching
                              -> interrupts
                              -> garbageCollection

    [Class variables:]

        InternalErrorHandler            gets informed (by VM), when some runtime
                                        error occurs (usually fatal)

        UserInterruptHandler            gets informed (by VM) when CNTL-C is pressed
        TimerInterruptHandler           gets alarm timer interrupts (from VM)
        SpyInterruptHandler             another alarm timer (from VM)
        StepInterruptHandler            gets single step interrupts (from VM)
        ExceptionInterruptHandler       gets floating point exceptions (from VM)
        ErrorInterruptHandler           gets primitive errors (from VM)
        MemoryInterruptHandler          gets soon-out-of-memory conditions (from VM)
        SignalInterruptHandler          gets unix signals (from VM)
        ChildSignalInterruptHandler     gets child death signals (from VM)
        DisposeInterruptHandler         gets informed, when an object is disposed from 
                                        a shadowArray (from VM)
        RecursionInterruptHandler       gets recursion limit violations (from VM)
        IOInterruptHandler              gets SIGIO unix signals (from VM)
        CustomInterruptHandler          gets custom interrupts (from VM)

        InterruptLatencyMonitor         if nonNil, that one will be notified (by the VM)
                                        with an interruptLatency:millis message for every 
                                        interrupt and gets the delay time
                                        (between the time when the signal arrived and
                                         when it was really delivered) as argument.
                                        This can be used to create a statistic for
                                        realtime systems.

        RegisteredErrorInterruptHandlers
                                        associates errorID (as passed from primitive
                                        to the __errorInterruptWithID() function)
                                        with handlers.

        IncrementalGCLimit              number of bytes, that must be allocated since
                                        last full garbage collect to turn the incremental
                                        collector on (at idle time).

        FreeSpaceGCLimit                low limit on freeSpace at which incremental
                                        gc starts to run at idle time.

        FreeSpaceGCAmount               amount to allocate once freeSpace drops
                                        below FreeSpaceGCLimit

        Dependents                      keep my dependents locally (its faster) for
                                        all those registries

        LowSpaceSemaphore               a semaphore signalled whenever the system is
                                        running in low memory (i.e. the memory manager
                                        ran into memory shortage and feels that it
                                        may soon no longer be able to grant allocation requests).
                                        You can have a process waiting on this semaphore
                                        which starts to remove (i.e. nil-out) objects
                                        or preform other cleanup actions.
                                        
        AllocationFailureSignal         signal raised when a new fails (see Behavior)
                                        When this signal is raised, the meomory manager
                                        is really in trouble (i.e. above feelings where
                                        correct)

        BackgroundCollectProcess        created by startBackgroundCollectorAt:

        BackgroundFinalizationProcess   created by startBackgroundFinalizationAt:

        ImageName                       name of the current image (or nil)

        ImageSaveTime   <AbsoluteTime>  timestamp when this image was saved


    Warning:
      The InterruptHandler variables are known by the runtime system -
      they are the objects that get an interrupt message when the event
      occurs. You may not remove any of them.

    [author:]
        Claus Gittinger

    [see also:]
        ( garbage collection : html:programming/GC.html )
"
!

garbageCollection
"
    Currently, Smalltalk/X uses a two-level memory hierachy (actually, there
    are more memory regions used for stack, permanent objects, symbols etc.
    but for the following discussion, these are not of interest).

  newSpace:

    Objects are created in a so-called newSpace, which is relatively small.
    This newSpace is cleaned by a so called ``scavenge''-operation, whenever 
    becoming full. Scavenging means, that all still-live objects (i.e. referenced
    by some other) are copied over to another memory area, leaving all unreachable
    objects as garbage behind. Thus, the newSpace actually consists of two semispaces,
    of whih only one is active - the other being used only while objects are
    copied.
    After this copying, these two semispaces exchange their roles - i.e. reachable
    objects are copied ping-pong like between these semispaces.
    Once an object survives enough of these copying operations, the next scavenge 
    will move it into the so called oldSpace, which is much larger, and not
    processed by the scavenger. 
    This movement of an object from newSpace to oldSpace is called ``tenure''.
    (this avoids objects being copied around forever).
    Once tenured, an object is no longer contained in the newSpace, and
    thus ceases to create any scavenging overhead after that.

    Scavenging occurs automatically, and is usually done fast enough to go 
    unnoticed (typically, it takes some 5 to 40ms to perform a scavenge, 
    depending on how many live objects are in the newspace).

    Interestingly, the scavenger performs better, if many garbage objects
    are to be reclaimed, since less object-copying has to be done. Therefore,
    the best-case scavenge time is almost zero, if there is only garbage in
    the newSpace. In contrast, the worst-case is when all newSpace objects are still
    living. Thus, from a newSpace collectors viewPoint, it makes sense to get
    objects out of the way as fast as possible. However the oldSpace is
    collected much less frequently and the cost to reclaim an oldspace object
    is much higher (actually, the cost to reclaim a newspace object is zero -
    its the survival of objects which we have to pay for).
    Therefore, from an oldSpace collectors point of view, its preferable to
    keep objects in the newSpace as long as possible.

    To honor this conflicting situation, the system uses an adaptive tenure-count,
    which adjusts the number of scavenges required for tenure (the so called 
    'tenureAge') according to the fill-grade of the newSpace.
    If the newSpace is relatively empty, it tries to keep objects longer there.
    The controlling parameters of the tenure age can be changed dynamically,
    detailed information is available upon request.

    The exact speed of the scavenger depends mostly on the speed of your memory
    interface (and, since most of todays memories have access times in the order
    of 50-100ns, the raw CPU speed does not correlate linear with the GC speed).
    Measurements give roughly 40ms for a full 400k newSpace 
    (i.e. all objects survive) on a 486/50 - this only drops to some 20-30ms on a P5.

    The upper bounds of the scavenge blocking time can be controlled by changing
    the size of the newSpace - either via a command line argument, or even dynamically
    by Objectmemory>>newSpaceSize:. Smaller sizes lead to shorter blocking periods,
    but greater absolute GC overhead. The default (400k) seems to be a good compromise.
    (if you are not happy with it, try playing around with the settings)

  oldSpace:

    To reclaim oldspace, the system uses three algorithms: mark&sweep, a copying
    (and compressing) baker-type collector and an incremental mark&sweep.

    The mark&sweep runs whenever the oldspace becomes full, putting dead objects
    onto a free list. If a memory request cannot be served from this freelist,
    and the total size of objects on the freelist exceeds a threshold, the system
    will compress the oldspace to make the free-space into one big area.
    This compress is done by copying all live objects into a newly allocated
    area, and freeing the previous memory afterwards (baker collector).
    Since a compressing oldspace collect leads to a noticable pause of the system,
    the memory manager tries hard to avoid oldspace compression.
    (well, if enough real memory is available to hold both spaces in physical
     memory, the compress is actually pretty fast).

    The incremental mark&sweep runs in the background, whenever the system is idle
    (see ProcessorSceduler>>waitForEventOrTimeout), or alternatively as a low or high
    priority background process (see ObjectMemory>>startBackgroundCollector). 
    Like the normal mark&sweep, this incremental collector follows object references 
    and marks reachable objects on its way. However, this is done 'a few objects-at-a-time',
    to not disrupt the system noticably. Currently, there are some (theoretical) and in
    practice never occurring situations, in which the incremental GC still creates noticable
    delays. A current project is involved with this and a future version of ST/X (ST/X-RT)
    will be available which shows deterministic worst case behavior in its GC pauses 
    (this will be provided as an additional add-on option - certainly not for free ;-).

    Currently, incremental GC blockings are in the order of 10-70ms.
    There is one catch with low priority background IGC: if there is never any idle
    time available (i.e. all processes run all the time), it would never get a chance
    to do any collection work. To handle this case, a background IGC can also be started
    as a high priority process, which gives up the cpu (by delaying on the time) after
    every IGC step. A high priority background collector will always make progress
    and eventually finish a GC cycle. However, it may have more of an influence on 
    the other processes. So, its up to you, to decide ...


    Incremental garbage collection is controlled by the variables 
    'IncrementalGCLimit', 'FreeSpaceGCLimit' and 'FreeSpaceGCAmount':

      the ProcessorScheduler will perform incremental GC steps at idle time, 
      if the total space allocated since the last full collect exceeds 
      'IncrementalGCLimit',
      or if there are less than 'FreeSpaceGCLimit' bytes available in free store.
      If after the incrementalGC, less than 'FreeSpaceGCLimi't bytes are available,
      'FreeSpaceGCAmount' more bytes are requested from the memory manager.

    The defaults are set in ObjectMemory>>initialize and can be changed in your 
    startup 'smalltalk.rc'-file. Setting them to nil turns incremental GC off.


    For example, setting 'IncrementalGCLimit' to 500000 will start the background collector
    whenever 500k bytes have been allocated - usually very seldom. Setting it to some
    small number (say 10000) will have it run very often.

    Setting 'FreeSpaceGCAmount' to (say) 1meg lets the system try to always keep
    1meg of freeSpace. If less memory is available, more oldSpace will be allocated.
    Keeping some memory in the pocket may prevent the system from running into a blocking
    GC if memory is allocated in peaks (but only, if the incremental GC can keep up with 
    allocation rate). The trigger level 'FreeSpaceGCLimit' should be below that amount;
    to avoid excessive incremental GC activity (say 1/4 if the amount).

    Having the background GC running often should not hurt the performance of your 
    smalltalk processes, since the IGC only runs at times when no ST processes are runnable. 
    (there are some short delays in event processing, since the IGC's steps may take 
    some XX ms.) 
    However, if you are not alone on your machine (i.e. a timesharing system) or 
    you have other Unix processes to run, you should not run the IGC too often, 
    since it may hurt OTHER users/unix processes.

    Since this collector only runs at idle times, even a low priority background 
    process will prevent it from doing its work. You may want to start a somewhat
    higher priority background collect (say at prio 4), which also preempts these
    background processes. (see ObjectMemory>>startBackgroundCollectorAt:).

    Beginning with 2.10.4, a third space, called symSpace has been added.
    Objects in this space are never moved or garbage collected.
    This space is used for (some) symbols only.

    Beginning with 2.10.5, a fourth space, called fixSpace has been added.
    Objects in this space are never moved or garbage collected.
    This space is used for constant objects (true, false, some basic classes etc.).

    A plan for 2.11 is to offer an arbitrary number of spaces, which can be
    attached and detached at runtime. This will allow easy share of object
    with remote systems and separating objects into a per application/package
    space. (be prepared for changes in the future and make your application
    independ of the VM internals)

  default setup:

    The following table lists some default settings and means for changing them:

            what        default     change by           
                                    command line arg    dynamically
    -----------------------------------------------------------------------
        newSpace size     400k      -Mnew nnn           newSpaceSize:nnn

        oldSpace size    3000k      -Mold nnn           moreOldSpace:
                                                        announceSpaceNeed:
                                                        collectGarbage

        max tenure age     29                           lockTenure:
                                                        avoidTenure:
                                                        (sets it to infinity)
        
        adaptive tenure                                 tenureParameters

        oldSpace
        compressor      enabled     -Msingle            -

        limit for
        old-compress     8000k      -                   oldSpaceCompressLimit:

        chunk size
        to increase
        oldpPace          256k      -                   oldSpaceIncrement:
        
        prefer moreOld
        to doing GC      false      -                   fastMoreOldSpaceAllocation:

        limit for
        above                -      -                   fastMoreOldSpaceLimit:

        keep size for        -      -                   freeSpaceGCAmount:
        IGC

        low water
        trigger for IGC      -      -                   freeSpaceGCLimit:

        allocated
        trigger for IGC   500k      -                   incrementalGCLimit

        
    By default, no incremental GC is started by the system; however,
    the standard startup script starts a low prio background incremental GC process.
    You have to edit your startup files to change this. 
    A suggested configuration (used by the author and the default) is:

        ' keep 1meg in the pocket '

        ObjectMemory freeSpaceGCAmount:1000000. 

        ' start incrementalGC when freespace drops below 250k '
        ' or 500k of oldSpace has been allocated              '

        ObjectMemory freeSpaceGCLimit:250000.                 '
        ObjectMemory incrementalGCLimit:500000.               '

        ' collect as a background process (the default is: at idle times)
        ' this means that running cubes or other demo processes are suspended
        ' for the collect; change the prio to below 4 if you want them to continue

        ObjectMemory startBackgroundCollectorAt:5.            '
        ObjectMemory startBackgroundFinalizationAt:5.         '

        ' quickly allocate more space (i.e. avoid blocking collects)
        ' up to 8meg - then start to collect if more memory is needed.

        ObjectMemory fastMoreOldSpaceLimit:8*1024*1024.       '
        ObjectMemory fastMoreOldSpaceAllocation:true.         '

  hints & tricks:

    normally, there is no need to call for an explicit garbage collection, or
    modify the default parameters.
    The memory system should adapt reasonable and provide good performance 
    for a wide range of allocation patterns (see Example3 below for an exception).

    However, there may be situations, in which hints and/or explicit
    control over allocation can speedup your programs; but please:

      - if you think you have to play around with the memory policies,
        first check your program - you may find useless allocations
        or bad uses of collections. A typical error that is made is to
        create large collections using the #, (comma) concatenation method,
        which shows square behavior, since it allocates many, many temporary
        collections. Also, watch out for #copyWith:, #add: etc.
        All of these create a new collection. Remember, that most collections
        offer methods to preallocate some space; for example, 'Set new:' creates
        an empty set, but preallocates space to avoid resizing over and over.

        An especially bad performace dog is to use #add: on fix-size collection
        objects (such as Strings or Arrays), since in addition to allocating
        lots of garbage, a #become: operation is required for EACH element
        added. NEVER use Arrays for growing/shrinking data - use OrderedCollection
        instead. (if you really need an array, use asArray afterwards)

      - if you are going to allocate huge data structures, think about
        optimizing space. For example, if you allocate a million instances of
        some object, each added instance variable makes up 4Mb of additional 
        memory need.
        Also, for Byte-valued, Integer-valued and Float like objects, special
        collections are provided, which store their values directly inside (instead
        of a reference to the object). A FloatArray consisting of 1 million floats
        requires about 4mb of memory, while an Array of Floats requires 4mb for the
        references to the floats, PLUS 20Mb for the floats themself.

      - check if you really need fast access to all of these objects; you may
        try to only keep some subset in memory, and use binary storage or
        (if this is too slow) optimized store/retrieve methods and keep the bigger
        part in a file. 
        (How about a DiskArray class, which does this transparently ?
         See the FileText class for some ideas and something to start with ...)


    Hint / Example 1: 
      you are about to allocate a huge data structure, which is known to
      survive long. In this case, it is better to have these objects move into the
      oldspace sooner, to avoid the copying overhead during scavenges.

      To do this, you can call ObjectMemory>>tenure after allocation, which
      forces all new-objects immediately into the oldspace. 
      Make certain, that not to many (ideally no) short-living objects are in the
      newspace when doing this.

      Another alternative is to tell the system that all allocation should be
      done directly in the oldspace. This completely avoids the scavenging overhead
      for these objects. To do so, use ObjectMemory>>turnGarbageCollectorOff
      before the allocation, and ObjectMemory>>turnGarbageCollectorOn afterwards.
      Keep in mind, that do-loops may allocate block-objects and other temporaries,
      so there is a danger of making things worse due to having all those temporaries
      in the oldspace afterwards. (which is not a fatal situation, but will
      force the system to do an oldspace collect earlier, which may not be your
      intention).


   Hint / Example 2:
      you know in advance, that a certain (big) amount of memory will be needed.
      For example, the fileBrowser wants to show a huge file in its text-view.
      In this case, it is better to tell the memory system in advance, how much
      memory will be needed, since otherwise many compresses and reallocations will
      occur (the memory system will allocate additional memory in chunks of smaller
      256k pieces, if a compress failes. Thus, if you are going to allocate (say) 1Mb of 
      strings, it will perform 5 compressing GC's).

      This is done using ObjectMemory>>moreOldSpace: or ObjectMemory announceSpaceNeed:.
      In the above example, you would do 'ObjectMemory announceSpaceNeed:500000', which 
      avoids those annoying 5 compressing GC's.
      BTW: if you have other smalltalk processes (threads) running which should not be
      paused if possible, it is better to use #announceSpaceNeed. This tries to avoid 
      pausing in other processes and sometimes succeeds, while moreOldSpace will always 
      block the whole system for a while. However, there is no 'no-pause' guarantee.

      The amount of automatic increase (in case the oldSpace becomes full) is 256k by
      default. This number can be changed with ObjectMemory>>oldSpaceIncrement:.


    Hint / Example3:
      There are rare cases, when an explicit GC makes a difference: since
      object finalization is done at GC time, objects which keep operatingSystem
      resources may be finalized late. This is normally no problem, except if
      the system is running out of resources. For example, allocating new colors
      may fail if many colors have already been allocated in the past - even
      though these colors are actually free. The Depth8Image calls for an
      explicit GC, whenever it fails to allocate a color for a bitmap, to force
      finalization of free, but not yet finalized colors.
      The same is true for things like file descriptors, if fileStreams are not closed
      correctly.

    Hint 4:
      If you run in too small of physical memory, the incremental GC may have a
      bad effect on your working set: since it touches pages (which may otherwise
      not be needed at the moment, the operating system is forced to steal other
      (possibly more useful) pages from your set of incore pages.
      Although ST/X would be willing to fix this behavior (by telling the OS about
      its page requirements, many OS's do not listen. The SGI-Iris, For example ignores
      the madvice system calls - other systems do not implement it.
      The trouble is that the standard LRU paging strategy is exactly the worst for
      a program which sequentially scans its memory once in a single direction ...
      You may get better performance, if you turn off the incremental GC while
      processing a big data structure.


    Warning: many of the methods found here are not standard and may not even be available in
    future versions of ST/X. Use them only in very special situations or experiments.

    Let me know about additional special features you think are useful, and about
    special features you are using - this provides the feedback required to decide
    which methods are to be removed, kept or enhanced in future versions.
"
!

interrupts
"
    Handling of interrupts (i.e. unix-signals) is done via handler objects, which
    get a #XXXInterrupt-message sent. This is more flexible than (say) signalling
    a semaphore, since the handler-object may do anything to react on the signal
    (of course, it can also signal a semaphore to emulate the above behavior).

    Another reason for having handler objects is that they allow interrupt handling
    without any context switch, for high speed interrupt response.
    However, if you do this, special care is needed, since it is not defined, 
    which (smalltalk-)process gets the interrupt and will do the processing 
    (therefore, the default setup installs handlers which simply signal a semaphore 
     and continue the current process).

    Typically, the handlers are set during early initialization of the system
    by sending 'ObjectMemory XXXInterruptHandler:aHandler' and not changed later.
    (see Smalltalk>>initialize or ProcessorScheduler>>initialize).
    To setup your own handler, create some object which responds to #xxxInterrupt,
    and make it the handler using the above method.

    Interrupt messages sent to handlers are:
        internalError:<someString>      - internal interpreter/GC errors
        userInterrupt                   - ^C interrupt
        customInterrupt                 - custom interrupt
        ioInterrupt                     - SIGIO interrupt
        timerInterrupt                  - alarm timer (SIGALRM)
        errorInterrupt:<id>             - errors from other primitives/subsystems
                                          (DisplayError)
        spyInterrupt                    - spy timer interrupt (SIGVTALARM)
        stepInterrupt                   - single step interrupt
        disposeInterrupt                - finalization required
        recursionInterrupt              - recursion (stack) overflow
        memoryInterrupt                 - soon running out of memory
        fpExceptionInterrupt            - floating point exception (SIGFPE)
        childSignalInterrupt            - death of a child process (SIGCHILD)
        signalInterrupt:<number>        - unix signal (if other than above signals)

    To avoid frustration in case of badly set handlers, these messages
    are also implemented in the Object class - thus anything can be defined
    as interrupt handler. However, the VM will not send any
    interrupt message, if the corresonding handler object is nil
    (which means that nil is a bad choice, if you are interested in the event).

    Interrupt processing is not done immediately after the event arrives: 
    there are certain ``save-places'' at which this handling is performed
    (message send, method return and loop-heads).
    If not explicitely enabled, primitive code is never interrupted.
    However, if you do enable interrupts in your primitive (see ExternalStream as example),
    be prepared for your objects to move around ... therefore, these have to
    be coded very carefully.

    Interrupts may be disabled completely (OperatingSystem blockInterrupts) and 
    reenabled (unblockInterrupts) to allow for critical data to be manipulated.
    The above are low-level primitive entries - you better use #valueUninterruptably,
    which cares for unwinds, long returns and restores the blocking state.

    Every process has its own interrupt-enable state which is switched
    when processes switch control (i.e. you cannot block interrupts across
    a suspend, delay etc.). 
    However, the state will be restored after a resume.
"
! !

!ObjectMemory class methodsFor:'initialization'!

initialize
    "initialize the class"

    AllocationFailureSignal isNil ifTrue:[
	AllocationFailureSignal := ErrorSignal newSignalMayProceed:true.
	AllocationFailureSignal nameClass:self message:#allocationFailureSignal.
	AllocationFailureSignal notifierString:'allocation failure'.

	MallocFailureSignal := AllocationFailureSignal newSignalMayProceed:true.
	MallocFailureSignal nameClass:self message:#mallocFailureSignal.
	MallocFailureSignal notifierString:'(malloc) allocation failure'.

	LowSpaceSemaphore := Semaphore new.
    ].
    DisposeInterruptHandler := self.
    IncrementalGCLimit := 500000.
    FreeSpaceGCLimit := FreeSpaceGCAmount := nil.
    MemoryInterruptHandler := self
! !

!ObjectMemory class methodsFor:'ST-80 compatibility'!

availableFreeBytes
    ^ self freeSpace + self freeListSpace

    "
     ObjectMemory availableFreeBytes 
    "
!

bytesPerOOP
    "return the number of bytes an object reference (for example: an instvar)
     takes"

%{  /* NOCONTEXT */
    RETURN(__MKSMALLINT(sizeof(OBJ)));
%}

    "
     ObjectMemory bytesPerOOP 
    "
!

bytesPerOTE
    "return the number of overhead bytes of an object.
     i.e. the number of bytes in every objects header."

%{  /* NOCONTEXT */
    RETURN(__MKSMALLINT(OHDR_SIZE));
%}

    "
     ObjectMemory bytesPerOTE 
    "
!

compactingGC
    self garbageCollect
!

current
    ^ self
!

globalCompactingGC
    self garbageCollect
!

growMemoryBy:numberOfBytes
    ^ self moreOldSpace:numberOfBytes
!

numOopsNumBytes
    ^ Array with:(self numberOfObjects)
	    with:(self bytesUsed)

    "
     ObjectMemory numOopsNumBytes 
    "
! !

!ObjectMemory class methodsFor:'Signal constants'!

allocationFailureSignal
    "return the signal raised when an object allocation failed"

    ^ AllocationFailureSignal
!

mallocFailureSignal
    "return the signal raised when malloc memory allocation failed.
     (usually, this kind of memory is used with I/O buffers or other temporary
      non-Object storage)"

    ^ MallocFailureSignal
! !

!ObjectMemory class methodsFor:'VM messages'!

debugPrinting
    "return true, if various debug printouts in the VM
     are turned on, false of off."

%{  /* NOCONTEXT */
    extern int __getDebugPrinting();

    RETURN (__getDebugPrinting() ? true : false);
%}
!

debugPrinting:aBoolean
    "turn on/off various debug printouts in the VM
     in case of an error. For example, a double-notUnderstood
     leads to a VM context dump if debugPrinting is on.
     If off, those messages are suppressed.
     The default is on, since these messages are only printed for
     severe errors."

%{  /* NOCONTEXT */
    __setDebugPrinting( (aBoolean == true) );
%}
!

infoPrinting
    "return true, if various informational printouts in the VM
     are turned on, false of off."

%{  /* NOCONTEXT */
    extern int __getInfoPrinting();

    RETURN (__getInfoPrinting() ? true : false);
%}
!

infoPrinting:aBoolean
    "turn on/off various informational printouts in the VM.
     For example, the GC activity messages are controlled by
     this flags setting.
     The default is true, since (currently) those messages
     are useful for ST/X developers."

%{  /* NOCONTEXT */
    __setInfoPrinting( (aBoolean == true) );
%}
! !

!ObjectMemory class methodsFor:'access debugging'!

setTrapOnAccessFor:anObject
    "install an access trap for anObject;
     An accessSignal will be raised, whenever any instvar of anObject is either read or written.
     This is not supported on all architectures, therefore the return value
     (true of trap was installed ok, false if failed) should be checked."

%{  /* NOCONTEXT */
    RETURN (__addTrapOnAccess(anObject, 2) ? true : false);
%}
!

setTrapOnReadFor:anObject
    "install a read trap for anObject;
     An accessSignal will be raised, whenever any access into anObject occurrs.
     This is not supported on all architectures, therefore the return value
     (true of trap was installed ok, false if failed) should be checked."

%{  /* NOCONTEXT */
    RETURN (__addTrapOnAccess(anObject, 0) ? true : false);
%}
!

setTrapOnWriteFor:anObject
    "install a write trap for anObject;
     An accessSignal will be raised, whenever any instvar of anObject is written to.
     This is not supported on all architectures, therefore the return value
     (true of trap was installed ok, false if failed) should be checked."

%{  /* NOCONTEXT */
    RETURN (__addTrapOnAccess(anObject, 1) ? true : false);
%}
!

unsetAllTraps
    "remove all access traps"

%{  /* NOCONTEXT */
    __removeAllAccessTraps();
%} 
!

unsetTrapFor:anObject
    "remove any access trap for anObject."

%{  /* NOCONTEXT */
    __removeTrapOnAccess(anObject);
%}
! !

!ObjectMemory class methodsFor:'cache management'!

flushCaches
    "flush method and inline caches for all classes"

%{  /* NOCONTEXT */
    __flushMethodCache();
    __flushAllInlineCaches();
%}
!

flushCachesFor:aClass
    "flush method and inline caches for aClass"

%{  /* NOCONTEXT */
    __flushMethodCacheFor(aClass);
    __flushInlineCachesFor(aClass);
%}
!

flushCachesForSelector:aSelector
    "flush method and inline caches for aSelector"

%{  /* NOCONTEXT */
    __flushMethodCacheForSelector(aSelector);
    __flushInlineCachesForSelector(aSelector);
%}
!

flushInlineCaches
    "flush all inlinecaches"

%{  /* NOCONTEXT */
    __flushAllInlineCaches();
%}
!

flushInlineCachesFor:aClass withArgs:nargs
    "flush inlinecaches for calls to aClass with nargs arguments"

%{  /* NOCONTEXT */
    __flushInlineCachesForAndNargs(aClass, __intVal(nargs));
%}
!

flushInlineCachesForClass:aClass
    "flush inlinecaches for calls to aClass."

%{  /* NOCONTEXT */
    __flushInlineCachesFor(aClass);
%}
!

flushInlineCachesForSelector:aSelector
    "flush inlinecaches for sends of aSelector"

%{  /* NOCONTEXT */
    __flushInlineCachesForSelector(aSelector);
%}
!

flushInlineCachesWithArgs:nargs
    "flush inlinecaches for calls with nargs arguments"

%{  /* NOCONTEXT */
    __flushInlineCaches(__intVal(nargs));
%}
!

flushMethodCache
    "flush the method cache"

%{  /* NOCONTEXT */
    __flushMethodCache();
%}
!

flushMethodCacheFor:aClass
    "flush the method cache for sends to aClass"

%{  /* NOCONTEXT */
    __flushMethodCacheFor(aClass);
%}
!

flushMethodCacheForSelector:aSelector
    "flush the method cache for sends of aSelector"

%{  /* NOCONTEXT */
    __flushMethodCacheForSelector(aSelector);
%}
!

incrementSnapshotID
%{  /* NOCONTEXT */
    __bumpSnapshotID();
%}
!

snapshotID
    "return the internal snapshotID number.
     This is incremented when an image is restarted, and
     stored with the image.
     Not for normal users, this is used by the VM to invalidate
     caches which are stored with the image"

%{  /* NOCONTEXT */
    RETURN ( __MKSMALLINT( __snapshotID() ));
%}
    "
     ObjectMemory snapshotID     
    "
!

trapRestrictedMethods:trap
    "Allow/Deny execution of restricted Methods (see Method>>>restricted:)

     Notice: method restriction is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    |oldTrap|

%{ 
	if (__setTrapRestrictedMethods(trap == true))
	    oldTrap = true;
	else
	    oldTrap = false;
%}.

    (trap and:[oldTrap not]) ifTrue:[
	self flushCaches
    ].
    ^ oldTrap

    "
	ObjectMemory trapRestrictedMethods:true
	ObjectMemory trapRestrictedMethods:false
    "
! !

!ObjectMemory class methodsFor:'debug queries'!

addressOf:anObject
    "return the core address of anObject as an integer
     - since objects may move around, the returned value is invalid after the
     next scavenge/collect.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( nil );
    }
    if ((unsigned)anObject <= _MAX_INT) {
	RETURN ( __MKSMALLINT((int)anObject) );
    }
    RETURN ( __MKUINT((int)anObject) );
%}
    "
    |p|
    p := Point new.
    ((ObjectMemory addressOf:p) printStringRadix:16) printNL.
    ObjectMemory scavenge.
    ((ObjectMemory addressOf:p) printStringRadix:16) printNL.
    "
!

ageOf:anObject
    "return the number of scavenges, an object has survived
     in new space. 
     For old objects and living contexts, the returned number is invalid.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( 0 );
    }
    RETURN ( __MKSMALLINT( _GET_AGE(anObject) ) );
%}
    "
    |p|
    p := Point new.
    (ObjectMemory ageOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    "
!

dumpObject:someObject
    "low level dump an object.
     WARNING: this method is for ST/X debugging only 
	      it may be removed (or replaced by a noop) without notice"

%{
    __dumpObject__(someObject);
%}
    "
     ObjectMemory dumpObject:true
     ObjectMemory dumpObject:(Array new:10) 
     ObjectMemory dumpObject:(10@20 corner:30@40) 
    "
!

flagsOf:anObject
    "For debugging only.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( nil );
    }
    RETURN ( __MKSMALLINT( anObject->o_flags ) );
%}
    "
F_ISREMEMBERED  1       /* a new-space thing being refd by some oldSpace thing */
F_ISFORWARDED   2       /* a forwarded object (you will never see this here) */
F_DEREFERENCED  4       /* a collection after grow (not currently used) */
F_ISONLIFOLIST  8       /* a non-lifo-context-referencing-obj already on list */
F_MARK          16      /* mark bit for background collector */
    "

    "
     |arr|

     arr := Array new.
     arr at:1 put:([thisContext] value).
     (ObjectMemory flagsOf:anObject) printNL
    "
!

globalRefChainTo:anObject
     |levels objects setOfOwners indent|

     levels := OrderedCollection new.
     objects := WeakIdentitySet with:anObject.

     [objects includes:Smalltalk] whileFalse:[
         setOfOwners := WeakIdentitySet new.

        objects do:[:o |
           |owners|

           owners := o allOwners.
           owners notNil ifTrue:[
               owners := owners select:[:el | (el isMemberOf:WeakArray) not].
               setOfOwners addAll:owners.
           ].

"/           (no instVarNamed:#keyArray) atAllPut:nil.
"/           no := nil.
"/           (owners instVarNamed:#keyArray) atAllPut:nil.
"/           owners := nil.
        ].

        levels add:objects.
'done with level: ' print. levels size print. ' found ' print. objects size print. ' refs' printNL.

        levels do:[:alreadyFound |
            setOfOwners remove:(alreadyFound instVarNamed:#keyArray) ifAbsent:nil.
            alreadyFound do:[:o |
                setOfOwners remove:o ifAbsent:nil
            ]
        ].

        setOfOwners remove:(objects instVarNamed:#keyArray) ifAbsent:nil.
"/        setOfOwners printNL.

        objects := setOfOwners.
        objects isEmpty ifTrue:[
"/            'none found' printNL.
            ^ nil
        ].
     ].

"/     '' printNL.

     ^ self 
        refChainFrom:Smalltalk
        to:anObject 
        inRefSets:levels reverse
        startingAt:1 

     "
      self globalRefChainTo:Point new
     "
     "
      Smalltalk at:#foo put:Point new.
      self globalRefChainTo:(Smalltalk at:#foo)
     "
     "
      |p|
      p := Point new.
      Smalltalk at:#foo put:(#x -> p).
      self globalRefChainTo:p
     "
     "
      |a|
      a := Array new:1. a at:1 put:a.
      Smalltalk at:#foo put:(#x -> a).
      self globalRefChainTo:a 
     "
     "
      |a|
      a := Array new:1. a at:1 put:a.
      Smalltalk at:#foo put:(#x -> (Array with:a)).
      self globalRefChainTo:a 
     "
     "
      |a|
      a := Array new:1. a at:1 put:a.
      Smalltalk at:#foo put:(#x -> (Array with:a with:(#y -> a))).
      self globalRefChainTo:a 
     "
     "
      |p|
      p := Point new.
      Smalltalk at:#foo put:(WeakArray with:(#x -> (#y -> p))).
      self globalRefChainTo:p
     "
     "
      self globalRefChainTo:Transcript topView
     "

    "Modified: 10.4.1996 / 20:15:16 / cg"
!

objectAt:anAddress
    "return whatever anAddress points to as object.
     BIG BIG DANGER ALERT: 
	this method is only to be used for debugging ST/X itself 
	- you can easily (and badly) crash the system.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

    |low high|

    low := anAddress bitAnd:16rFFFF.
    high := (anAddress bitShift:16) bitAnd:16rFFFF.
%{
    if (__bothSmallInteger(low, high)) {
	RETURN ((OBJ)((__intVal(high) << 16) | __intVal(low)));
    }
%}
!

nteger(low, high)) {
	RETURN ((OBJ)((__intVal(high) << 16) | __intVal(low)));
    }
%}
!

oReferences: or anObject>>allOwners."

%{
    __printRefChain(__context, anObject);
%}
!

to:o inRefSets:levels startingAt:1. 
     "

    "Modified: 19.3.1996 / 23:22:41 / cg"
!

hist at:nw put:(hist at:nw) + 1
	 ].
     ].
     hist printNL.
     big printNL
    "
! !

!ObjectMemory class methodsFor:'dependents access'!

INT( __qSpace(anObject) ) );
%}
!

ts access'!

endents:aCollection
    "set the dependents collection"

    Dependents := aCollection
! !

!ObjectMemory class methodsFor:'enumerating'!

[
		aBlock value:each
	    ]
	]
    ]
!

esOfDo(&aClass, &aBlock, &work COMMA_CON) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

sOfDo((OBJ *)0, &aBlock, &work COMMA_CON) < 0) {
	RETURN (false);
    }
%}.
    ^ true
! !

!ObjectMemory class methodsFor:'garbage collection'!

(false);
    }
%}. 
    ^ true
!

kgroundCollectProcess notNil

    "
     ObjectMemory backgroundCollectorRunning
    "
!

AndSweep(__context);
    }
%}

    "
     ObjectMemory compressingGarbageCollect
    "
!

t to markAndSweep

    self markAndSweep

    "
     ObjectMemory garbageCollect
    "
!

extern int __incrGCstep();

    RETURN (__incrGCstep(__context) ? true : false);
%}
!

OldSpaceIfUseful
		].
	    ] ifFalse:[
		done := true
	    ]
	].
    ].
    ^ done not
!

ObjectMemory incrementalGC] forkAt:3
     [ObjectMemory incrementalGC] forkAt:9 
    "
!

ble."
%{
    __markAndSweep(__context);
%}

    "
     ObjectMemory markAndSweep
    "
!

."
%{
    __reclaimSymbols(__context);
%}
    "
     ObjectMemory reclaimSymbols
    "
!

."
%{
    __nonTenuringScavenge(__context);
%}

    "
     ObjectMemory scavenge
    "
!

tMemory freeSpaceGCLimit:1000000.
     ObjectMemory startBackgroundCollectorAt:5
    "
!

oundCollectProcess := nil
    ]

    "
     ObjectMemory stopBackgroundCollector
    "
!

tMemory tenure
     ... continue - objects created above are now in oldSpace ...
    "
!

ect)"
%{
    __scavenge(__context);
%}

    "
     ObjectMemory tenuringScavenge
    "
! !

!ObjectMemory class methodsFor:'garbage collector control'!

oseGarbageCollect
    "
!

eOldSpace:howMuch
	]
    ]

    "
     ObjectMemory announceOldSpaceNeed:1000000
    "
!

moreOldSpace:howMuch
	]
    ]

    "
     ObjectMemory announceSpaceNeed:100000
    "
!

ERIMENTAL interface."

%{  /* NOCONTEXT */
    __avoidTenure(flag == true ? 1 : 0);
%}
!

!

eAllocation:true.
     ObjectMemory fastMoreOldSpaceAllocation:prev.
     ^ prev
    "
!

aceLimit:10*1024*1024.
     ObjectMemory fastMoreOldSpaceLimit:prev.
     ^ prev
    "
!

riate value"

    ^ FreeSpaceGCAmount

    "
     ObjectMemory freeSpaceGCAmount
    "
!

ppropriate amount ...)
    "
    "
     ObjectMemory freeSpaceGCAmount:nil.     
    "
!

trigger off."

    ^ FreeSpaceGCLimit

    "
     ObjectMemory freeSpaceGCLimit
    "
!

ystem hit the wall ...)
    "
    "
     ObjectMemory freeSpaceGCLimit:nil.     
    "
!

gger off."

    ^ IncrementalGCLimit

    "
     ObjectMemory incrementalGCLimit
    "
!

'do incr. GC very often'
     ObjectMemory incrementalGCLimit:nil.     'never'
    "
!

flag == true) {
	__tenure(__context);
    }
    __lockTenure(flag == true ? 1 : 0);
%}
!

if (__moveToOldSpace(anObject, __context) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

lse );
    }
    RETURN (false);
%}
    "
     ObjectMemory moreOldSpace:1000000
    "
!

ease oldSpace - reset limit' errorPrintNL.
		FreeSpaceGCLimit := nil
	    ]
	].
    ].
!

Size:200*1024
    "

    " the default:

     ObjectMemory newSpaceSize:400*1024
    "
!

T( __compressingGCLimit(-1) ));
%}
    "
     ObjectMemory oldSpaceCompressLimit
    "
!

e the limit to 12Mb:"
    "
     ObjectMemory oldSpaceCompressLimit:12*1024*1024
    "
!

LLINT( __oldSpaceIncrement(-1) )); 
%}
    "
     ObjectMemory oldSpaceIncrement
    "
!

o change increment to 1Meg:"
    "
     ObjectMemory oldSpaceIncrement:1024*1024
    "
!

e system may behave very strange."

%{  /* NOCONTEXT */
    __tenureParams(magic);
%}.
!

n by the time you read this."

%{  /* NOCONTEXT */
    __allocForceSpace(OLDSPACE);
%}
!

Memory>>turnGarbageCollectorOff)"

%{  /* NOCONTEXT */
    __allocForceSpace(9999);
%}
! !

!ObjectMemory class methodsFor:'interrupt handler access'!

lag == true ? 1 : 0);
%}
!

for UNIX-death-of-a-childprocess-signal interrupts"

    ^ ChildSignalInterruptHandler
!

reated: 22.12.1995 / 14:14:52 / stefan"
    "Modified: 22.12.1995 / 14:15:16 / stefan"
!

ptHandler
    "return the handler for custom interrupts"

    ^ CustomInterruptHandler
!

er
    "set the handler for custom interrupts"

    CustomInterruptHandler := aHandler
!

"return the handler for object disposal interrupts"

    ^ DisposeInterruptHandler
!

t the handler for object disposal interrupts"

    DisposeInterruptHandler := aHandler
!

ler
    "return the handler for display error interrupts"

    ^ ErrorInterruptHandler
!

"set the handler for display error interrupts"

    ErrorInterruptHandler := aHandler
!

the handler for floating point exception interrupts"

    ^ ExceptionInterruptHandler
!

c.  
     Those should not occur in normal circumstances."

    ^ InternalErrorHandler
!

handler for I/O available signal interrupts (SIGIO/SIGPOLL)"

    ^ IOInterruptHandler
!

r I/O available signal interrupts (SIGIO/SIGPOLL)"

    IOInterruptHandler := aHandler
!

the handler for recursion/stack overflow interrupts"

    ^ RecursionInterruptHandler
!

er for recursion/stack overflow interrupts"

    RecursionInterruptHandler := aHandler
!

ctionary new
    ].
    RegisteredErrorInterruptHandlers at:errorIDSymbol put:aHandler
!

rruptHandlers
    "return registered handlers"

    ^ RegisteredErrorInterruptHandlers
!

dler
    "return the handler for UNIX-signal interrupts"

    ^ SignalInterruptHandler
!

"set the handler for UNIX-signal interrupts"

    SignalInterruptHandler := aHandler
!

ptHandler
    "return the handler for spy-timer interrupts"

    ^ SpyInterruptHandler
!

er
    "set the handler for spy-timer interrupts"

    SpyInterruptHandler := aHandler
!

andler
    "return the handler for single step interrupts"

    ^ StepInterruptHandler
!

"set the handler for single step interrupts"

    StepInterruptHandler := aHandler
!

ruptHandler
    "return the handler for timer interrupts"

    ^ TimerInterruptHandler
!

dler
    "set the handler for timer interrupts"

    TimerInterruptHandler := aHandler
!

ler
    "return the handler for CNTL-C interrupt handling"

    ^ UserInterruptHandler
! !

!ObjectMemory class methodsFor:'interrupt statistics'!

InterruptHandler := aHandler
!

"

    "Created: 7.11.1995 / 21:05:50 / cg"
    "Modified: 7.11.1995 / 21:13:33 / cg"
!

ptLatencyMonitor := self.
    ]

    "
     ObjectMemory interruptLatencyGoal:50
    "
!

non-standard debugging/realtime instrumentation entry."

    ^ InterruptLatencyMonitor
!

rd debugging/realtime instrumentation entry."

    InterruptLatencyMonitor := aHandler
!

ccumulated interrupt latency in millis.
     DEMO Example."

    ^ MaxInterruptLatency
! !

!ObjectMemory class methodsFor:'low memory handling'!

MaxInterruptLatency := 0
!

orPrintNL.

    LowSpaceSemaphore signalIf.

    "Modified: 12.4.1996 / 14:57:55 / cg"
! !

!ObjectMemory class methodsFor:'object finalization'!

d: 12.4.1996 / 14:57:28 / cg"
!

lost a pointer recently."
%{
    __allChangedShadowObjectsDo(&aBlock COMMA_CON);
%}
!

tionSemaphore signal
    ] ifFalse:[
	"/
	"/ do it right here
	"/
	self finalize
    ]
!

self allChangedShadowObjectsDo:[:aShadowArray | 
	aShadowArray lostPointer.
    ]
!

inalizationProcess := p

    "
     ObjectMemory startBackgroundFinalizationAt:5
    "
! !

!ObjectMemory class methodsFor:'physical memory access'!

ckgroundFinalization
    "
!

llectedOldSpacePagesDo(&aBlock COMMA_CON) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

if (__newSpacePagesDo(&aBlock COMMA_CON) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

if (__oldSpacePagesDo(&aBlock COMMA_CON) < 0) {
	RETURN (false);
    }
%}.
    ^ true
! !

!ObjectMemory class methodsFor:'queries'!

supply this info - assume yes"
    ^ true
!

__newSpaceUsed() - __freeListSpace()) );
%}
    "
     ObjectMemory bytesUsed  
    "
!

[
	"actually this cannot happen - there is always one"
	^ nil
    ].
    ^ aCollection
!

TURN ( __MKSMALLINT(__fixSpaceSize()) );
%}
    "
     ObjectMemory fixSpaceSize
    "
!

TURN ( __MKSMALLINT(__fixSpaceUsed()) );
%}
    "
     ObjectMemory fixSpaceUsed
    "
!

RN ( __MKSMALLINT(__freeListSpace()) );
%}
    "
     ObjectMemory freeListSpace
    "
!

INT(__oldSpaceSize() - __oldSpaceUsed()) );
%}
    "
     ObjectMemory freeSpace
    "
!

LLINT(__garbageCollectCount()));
%}
    "
     ObjectMemory garbageCollectCount 
    "
!

SMALLINT(__incrementalGCCount()));
%}
    "
     ObjectMemory incrementalGCCount
    "
!

EXT */
    extern int __incrGCphase();

    RETURN (__MKSMALLINT(__incrGCphase()));
%}
!

ctMemory lastScavengeReclamation)
      / (ObjectMemory newSpaceSize)) * 100.0  
    "
!

RETURN ( __lifoRememberedSet() );
%}
    "
     ObjectMemory lifoRememberedSet
    "
!

NT(__lifoRememberedSetSize()));
%}
    "
     ObjectMemory lifoRememberedSetSize
    "
!

KSMALLINT(__markAndSweepCount()));
%}
    "
     ObjectMemory markAndSweepCount 
    "
!

HASH__ << __HASH_SHIFT__) );
%}
    "
     ObjectMemory maximumIdentityHashValue
    "
!

T(__newSpaceReclaimedMin()) );
%}
    "
     ObjectMemory minScavengeReclamation
    "
!

TURN ( __MKSMALLINT(__newSpaceSize()) );
%}
    "
     ObjectMemory newSpaceSize
    "
!

N ( __MKSMALLINT(__newSpaceUsed()) );
%}
    "
     ObjectMemory newSpaceUsed   
    "
!

bj | tally := tally + 1].
    ^ tally

    "
     ObjectMemory numberOfObjects  
    "
!

__MKSMALLINT(__weakListSize()) );
%}
    "
     ObjectMemory numberOfWeakObjects
    "
!

atedSinceLastGC()) );
%}
    "
     ObjectMemory oldSpaceAllocatedSinceLastGC   
    "
!

TURN ( __MKSMALLINT(__oldSpaceSize()) );
%}
    "
     ObjectMemory oldSpaceSize
    "
!

RN ( __MKSMALLINT(__oldSpaceUsed()) );
%}
    "
     ObjectMemory oldSpaceUsed  
    "
!

MKSMALLINT(__rememberedSetSize()));
%}
    "
     ObjectMemory rememberedSetSize
    "
!

jectMemory resetMinScavengeReclamation.
     ObjectMemory minScavengeReclamation
    "
!

ngleOldSpace() ? true : false) );
%}
    "
     ObjectMemory runsSingleOldSpace 
    "
!

URN (__MKSMALLINT(__scavengeCount()));
%}
    "
     ObjectMemory scavengeCount 
    "
!

TURN ( __MKSMALLINT(__symSpaceSize()) );
%}
    "
     ObjectMemory symSpaceSize
    "
!

TURN ( __MKSMALLINT(__symSpaceUsed()) );
%}
    "
     ObjectMemory symSpaceUsed
    "
!

*/
    extern unsigned __tenureAge();

    RETURN ( __MKSMALLINT(__tenureAge()) );
%}
!

references:anObject]

    "
     (ObjectMemory whoReferences:Transcript) printNL
    "
!

:aClass]

    "
     (ObjectMemory whoReferencesDerivedInstancesOf:View) printNL
    "
! !

!ObjectMemory class methodsFor:'semaphore access'!

sOf:SystemBrowser) printNL
    "
! !

!ObjectMemory class methodsFor:'statistics'!

ease object."

    ^ LowSpaceSemaphore
! !

!ObjectMemory class methodsFor:'system management'!

T */

    __ageStatistics();
%}
!

ay
     change without notice."

%{
    __REGISTRATION_DO_BLOCK(&aBlock COMMA_SND);
%}
!

ObjectMemory binaryModuleInfo
    "

    "Modified: 17.9.1995 / 16:33:02 / claus"
!

ObjectMemory fullBinaryModuleInfo
    "

    "Modified: 30.8.1995 / 17:29:30 / claus"
!

nm copyWithoutLast:4
    ].
    ^ nm

    "
     ObjectMemory imageBaseName    
    "
!

not running from an image."

    ^ ImageName

    "
     ObjectMemory imageName 
    "
!

g image was saved.
     Return nil if not running from an image."

    ^ ImageSaveTime
!

tNil and:[newClass implements:#initialize]) ifTrue:[
	    newClass initialize
	]
    ]
!

"/    ^ self imageBaseName , '.chg'

    "
     ObjectMemory nameForChanges    
    "
!

ngeFileName := aFilename

    "
     ObjectMemory nameForChanges:'myChanges'    
    "
!

^ self imageBaseName , '.img'

    "
     ObjectMemory nameForSnapshot    
    "
!

"

    ^ self imageBaseName , '.src'

    "
     ObjectMemory nameForSources    
    "
!

n(__context, _stringVal(aFileName), funny);
	__UNBLOCKINTERRUPTS();
    }
%}.
    ^ ok
!

."

    self snapShotOn:(self nameForSnapshot)

    "
     ObjectMemory snapShot
    "
! !

!ObjectMemory class methodsFor:'documentation'!

mory snapShotOn:'myimage.img'
    "
! !
ObjectMemory initialize!
